from typing import List

from scholarly import scholarly
from scholarly.author_parser import AuthorParser
from scholarly.data_types import Author, Publication, CitesPerYear

from .utilities import getObjectPublicAttributes


class SimplifiedPublication:
    """
    :class:`SimplifiedPublication <SimplifiedPublication>` object used to represent a simplified publication
    entry on Google Scholar.
    """

    author_id: List[str] = None
    """
    List of the corresponding author ids of the authors that contributed to the Publication. 
    (source: PUBLICATION_SEARCH_SNIPPET)
    """

    num_citations: int = None
    "Number of citations of this Publication."

    cites_per_year: CitesPerYear = None
    "A dictionary containing the number of citations per year for this Publication. (source: AUTHOR_PUBLICATION_ENTRY)"

    pub_url: str = None
    "URL of the website providing the publication."

    author_pub_id: str = None
    """
    The id of the paper on Google Scholar from an author page. Comes from the
    parameter "citation_for_view=PA9La6oAAAAJ:YsMSGLbcyi4C". It combines the
    author id, together with a publication id. It may corresponds to a merging
    of multiple publications, and therefore may have multiple "citedby_id"
    values. (source: AUTHOR_PUBLICATION_ENTRY)
    """

    url_related_articles: str = None
    "The URL containing link for related articles of a publication. (needs fill() for AUTHOR_PUBLICATION_ENTRIES)"

    _class_attributes: List[str] = None
    "List of all the public attributes of the class."

    #
    # From there, the attributes are obtained from the BibEntry object : "bib" attribute of a publication.
    #
    pub_type: str = None
    "The type of entry for this bib (for example 'article'). (source: PUBLICATION_SEARCH_SNIPPET)"

    bib_id: str = None
    "Bib entry id. (source: PUBLICATION_SEARCH_SNIPPET)"

    abstract: str = None
    "Description of the publication."

    title: str = None
    "Title of the publication."

    author: str = None
    "List of the author names that contributed to this publication."

    pub_year: str = None
    "The year the publication was first published."

    venue: str = None
    "The venue of the publication. (source: PUBLICATION_SEARCH_SNIPPET)"

    journal: str = None
    "Journal Name. (source: PUBLICATION_SEARCH_SNIPPET)"

    volume: str = None
    "Number of years a publication has been circulated."

    number: str = None
    "NA number of a publication."

    pages: str = None
    "Range of pages."

    publisher: str = None
    "The publisher's name."

    citation: str = None
    """
    Formatted citation string, usually containing journal name, volume and page numbers.
    (source: AUTHOR_PUBLICATION_ENTRY)
    """

    def __init__(self, publication: Publication):
        # Get all the public attributes of the class
        self._class_attributes = getObjectPublicAttributes(self)

        if publication['container_type'] != "Publication":
            raise ValueError("The given object is not a Publication object.")

        if not publication['filled']:
            scholarly.fill(publication)

        # Remove the attributes that are defined by the "bib" attribute of the publication,
        # since it is a BibEntry object, and we want it right in the object
        attributes_to_use = self._class_attributes.copy()
        bib_attributes = None
        if 'bib' in publication:
            bib_attributes = list(publication['bib'])
            for attribute in bib_attributes:
                if attribute in attributes_to_use:
                    attributes_to_use.remove(attribute)

        # Copy all the needed attributes to the object
        for attribute in self._class_attributes:
            if attribute in publication:
                self.__dict__[attribute] = publication[attribute]

        # Handle the specificity of the attributes of the bib: Copy them to the root of the object
        if bib_attributes:
            for attribute in bib_attributes:
                if attribute in self._class_attributes:
                    self.__dict__[attribute] = publication['bib'][attribute]

    def __str__(self):
        """
        Simple string representation of the object.
        :return: The string representation of the object
        """
        return str(self.__dict__)


class SimplifiedCoauthor:
    """
    :class:`SimplifiedCoauthor <SimplifiedCoauthor>` object used to represent a simplified coauthor entry on Google Scholar.
    """

    scholar_id: str = None
    "The id of the author on Google Scholar."

    name: str = None
    "The name of the author."

    affiliation: str = None
    "The affiliation of the author."

    _class_attributes: List[str] = None
    "List of all the public attributes of the class."

    def __init__(self, coauthor: dict):
        # Get all the public attributes of the class
        self._class_attributes = getObjectPublicAttributes(self)

        if coauthor['container_type'] != "Author":
            raise ValueError("The given object is not an Author object.")

        # Copy all the needed attributes to the object
        for attribute in self._class_attributes:
            if attribute in coauthor:
                self.__dict__[attribute] = coauthor[attribute]

    def __str__(self):
        """
        Simple string representation of the object.
        :return: The string representation of the object
        """
        return str(self.__dict__)


class SimplifiedAuthor(SimplifiedCoauthor):
    """
    :class:`SimplifiedAuthor <SimplifiedAuthor>` object used to represent a simplified author entry on Google Scholar.
    """

    organization: int = None
    "A unique ID of the organization. (source: AUTHOR_PROFILE_PAGE)"

    homepage: str = None
    "URL of the homepage of the author."

    citedby: int = None
    "The number of citations to all publications. (source: SEARCH_AUTHOR_SNIPPETS)."

    interests: List[str] = None
    "Fields of interest of this Author. (sources: SEARCH_AUTHOR_SNIPPETS, AUTHOR_PROFILE_PAGE)"

    cites_per_year: CitesPerYear = None
    "Breakdown of the number of citations to all publications over the years. (source: SEARCH_AUTHOR_SNIPPETS)"

    publications: List[SimplifiedPublication] = None
    "A list of publications objects. (source: SEARCH_AUTHOR_SNIPPETS)"

    coauthors: List[SimplifiedCoauthor] = None  # List of authors. No self dict functionality available
    "A list of coauthors (list of Author objects). (source: SEARCH_AUTHOR_SNIPPETS)"

    def __init__(self, author: Author = None):
        # Call super constructor
        super().__init__(author)

        if author['container_type'] != "Author":
            raise ValueError("The given object is not an Author object.")

        if len(author['filled']) != len(AuthorParser(None)._sections):
            scholarly.fill(author)

        # Remove the attributes 'publications' and 'coauthors' from the list, since they are lists of
        # SimplifiedPublication and SimplifiedCoauthor objects. Make sure the original list is not modified.
        attributes_to_use = self._class_attributes.copy()
        attributes_to_use.remove('publications')
        attributes_to_use.remove('coauthors')

        # Copy all the needed attributes to the object
        for attribute in attributes_to_use:
            if attribute in author:
                self.__dict__[attribute] = author[attribute]

        # Casting the publications to SimplifiedPublication objects if they exist and are of type list
        if 'publications' in author and isinstance(author['publications'], list):
            self.publications = [SimplifiedPublication(pub) for pub in author['publications']]

        # Casting the coauthors to SimplifiedCoauthor objects if they exist and are of type list
        if 'coauthors' in author and isinstance(author['coauthors'], list):
            self.coauthors = [SimplifiedCoauthor(coauthor) for coauthor in author['coauthors']]

    def __str__(self):
        """
        Simple string representation of the object.
        :return: The string representation of the object
        """
        return str(self.__dict__)



import json
from queue import Queue
from threading import Thread
from typing import List

from scholarly import scholarly

from .CustomScholarlyTypes import SimplifiedAuthor
from .utilities import JSONEncoder


def getAuthorData(scholarId: str):
    """
    Retrieve the author's data from Google Scholar.
    :param scholarId: The id of the author on Google Scholar.
    :return: The author's data.
    """
    # Retrieve the author's data
    author = scholarly.search_author_id(scholarId)

    # Cast the author to Author object
    newauthor = SimplifiedAuthor(author)

    return newauthor


# Threaded function for queue processing.
def crawl(queue: Queue, results: List):
    """
    Crawl the author's data from Google Scholar.
    :param queue: The queue to fetch the work from.
    :param results: The list to append the results to.
    :return: Always true.
    """
    while not queue.empty():
        # Fetch new work from the Queue
        work = queue.get()

        try:
            data = getAuthorData(work[1])
            results.append(data)
        except:
            pass

        # Signal to the queue that task has been processed
        queue.task_done()
    return True


class ScholarScraper:
    """
    :class:`ScholarScraper <ScholarScraper>` object used to retrieve the data of a list of authors from Google Scholar.
    """

    def __init__(self, scholarIds: List[str] = [], max_threads: int = 10):
        """
        :param scholarIds: The list of the ids of the authors on Google Scholar.
        :param max_threads: The maximum number of threads to use for the scraping process.
        """
        self.scholarIds = scholarIds
        self.max_threads = max_threads
        self.authorsList = []
        self.threads = []

    def start_scraping(self, scholarIds: List[str] = None, max_threads: int = None):
        """
        Start the scraping process.
        :param scholarIds: The list of the ids of the authors on Google Scholar.
        :param max_threads: The maximum number of threads to use for the scraping process.
        :return: The list of the authors' data as JSON.
        """
        self.authorsList = []
        self.threads = []
        self.scholarIds = scholarIds if scholarIds else self.scholarIds
        self.max_threads = max_threads if max_threads else self.max_threads

        # Initialize an infinite queue
        queue = Queue(maxsize=0)

        # Use many threads (self.max_threads max, or one for each scholarId)
        num_theads = min(self.max_threads, len(self.scholarIds))

        # Iterate over all the scholarIds using index
        for index_id, scholarId in enumerate(self.scholarIds):
            queue.put((index_id, scholarId))

        # Starting worker threads on queue processing
        for i in range(num_theads):
            worker = Thread(target=crawl, args=(queue, self.authorsList))
            worker.start()

        # Now we wait until the queue has been processed
        queue.join()

        return json.dumps(self.authorsList, cls=JSONEncoder, sort_keys=True, indent=4, ensure_ascii=False)



from json import JSONEncoder


def getObjectPublicAttributes(obj):
    """
    Get all the public attributes of an object.
    :param obj: The object to get the attributes from.
    :return: The dictionary of the public attributes.
    """
    return [attr for attr in dir(obj.__class__) if
            not callable(getattr(obj.__class__, attr))
            and not attr.startswith("__")
            and not attr.startswith("_")
            and not attr.startswith("_" + obj.__class__.__name__ + "__")]


class JSONEncoder(JSONEncoder):
    """
    Simple JSON encoder class that allows to serialize objects that are not serializable by default.
    """

    def default(self, o):
        """
        Returns the object's dictionary if it has one But only the attributes that are returned by the
        :meth:`getObjectPublicAttributes` function.
        :param o: The object to serialize
        :return: The object's dictionary
        """
        if not hasattr(o, '__dict__'):
            return JSONEncoder.default(self, o)

        if hasattr(o, '_class_attributes'):
            return {k: v for k, v in o.__dict__.items() if k in o._class_attributes}

        return {k: v for k, v in o.__dict__.items() if k in getObjectPublicAttributes(o)}



from setuptools import setup, find_packages

setup(
    name='scholar-scraper',
    version='1.0.1',
    description='A python API to get publications from specific users on Google Scholar.',
    long_description_content_type="text/markdown",
    long_description=open('README.md').read(),
    url='https://github.com/guillaume-elambert/ScholarPythonAPI',
    keywords=['Google Scholar', 'academics', 'citations'],
    author='Guillaume ELAMBERT',
    author_email='guillaume.elambert@yahoo.fr',
    packages=find_packages(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development :: Libraries :: Python Modules'],
    install_requires=[
        'scholarly~=1.7.9',
    ],
    python_requires='>=3',
    license='GPLv3+',
)
